module clkRateTool32
  #(parameter CLKREF_RATE_IN_MHZ = 100)  // rate of 'clkref' in MHz
  (input 	     clkref,   // reference clock used to compare against clktest
   input 	     clktest, // clock to be measured
   input 	     clktest_en, // clock to be measured
   input 	     clkvalue, // clock used to sync the output value (can be clkref or clktest or different clock)
   output reg [31:0] value);
	 
   // maximum rate of clkref is 838 MHz (2^23). Can increase
   // REFCTR_MSB if need to use with a faster clkref.
   parameter REFCTR_SIZE = 23; 
   
   // **** clkref Clock Domain ****
   //
   reg [REFCTR_SIZE-1:0] refCtr;      
   reg [REFCTR_SIZE-1:0] rateCtr;
   reg 		         counting_clkref;

   // clktest_div8 is really in clktest clock domain 
   // but referenced in this clock domain
   reg 			 clktest_div1; 
   reg 			 clktest_div2; 
   reg 			 clktest_div4; 
   reg 			 clktest_div8; 

   reg 			 clktest_div8_ref1;      
   reg 			 clktest_div8_ref2;
   reg 			 clktest_div8_post;
   
   // number of counts of refCtr in 10 ms (minus 1)
   parameter REFCTR_TERMINAL = (((CLKREF_RATE_IN_MHZ*1000000)/100)-1);
   
   always @(posedge clkref) refCtr <= refCtr + 'b1;

   // The way this frequency counter basically works is to count the
   // number of rising edges seen in 10 ms. Once the 10 ms period is up,
   // allow this value to be copied over to the clkvalue clock domain.
   // The 10 ms counter is continuing to count during the copy but the
   // number of rising edges count remains static. Once the 10 ms counter
   // reaches its highest value, it resets to 0 and the 10 ms
   // time period restarts. By this point, it is expected that the number
   // of rising edges counter has been successfully copied.
   //
   // In order to count the clktest rising edges reliably, clkref must be
   // faster than clktest, so divide clktest down by 8. The loss of precision
   // is low compared to the being able to count clocks up to 4-8 times
   // clkref.
   
   // Count the number of rising & falling edges of clktest_div8 seen in 10 ms.
   // Using clktest/8 instead of clktest because sometimes connecting
   // a clock directly to logic can cause clock skew. 
   always @(posedge clkref) begin
      clktest_div8_ref1 <= clktest_div8;
      clktest_div8_ref2 <= clktest_div8_ref1;
      clktest_div8_post <= clktest_div8_ref2;

      if (refCtr == {REFCTR_SIZE{1'b1}}) begin
	 // At the largest count possible for refCtr, reset rateCtr
	 // and indicate that rateCtr is being modified by asserting
	 // counting_clkref.
	 counting_clkref <= 1;
	 rateCtr <= 0;
      end else if (refCtr <= REFCTR_TERMINAL) begin
	 // While within the 10 ms time, count every rising edge of clktest/2
	 counting_clkref <= 1;
	 if ((clktest_div8_ref2 && !clktest_div8_post) || (!clktest_div8_ref2 && clktest_div8_post))
	   rateCtr <= rateCtr + 'b1;
      end else begin
	 // Done with 10 ms timer, so deassert counting_clkref so
	 // other clock domain knows it can safely copy rateCtr, which
	 // is being kept unmodified.
	 counting_clkref <= 0;
	 rateCtr <= rateCtr;
      end	 

   end

   //
   // **** clkref Clock Domain ****
   

   // **** clktest Clock Domain ****
   //

   // Create a signal that is clktest/2 so can interface clktest to
   // logic without causing any possible clock skew
   always @(posedge clktest) if(clktest_en) begin
      clktest_div1 <= !clktest_div1;
      if (clktest_div1 == 'b1) begin
	clktest_div2 <= !clktest_div2;
      end
      if (clktest_div2 == 'b1 && clktest_div1 == 'b1) begin
	 clktest_div4 <= !clktest_div4;
      end
      if (clktest_div4 == 'b1 && clktest_div2 == 'b1 && clktest_div1 == 'b1) begin
	 clktest_div8 <= !clktest_div8;
      end
   end
   
   //
   // **** clktest Clock Domain ****

   // **** clkvalue Clock Domain ****
   //
   reg 		     counting_sync1;
   reg 		     counting_sync2;
   reg 		     counting;
   reg 		     counting_post;

   // sync counting_clkref to clkvalue. Use extra f/f just in case.
   always @(posedge clkvalue) begin
      counting_sync1 <= counting_clkref;
      counting_sync2 <= counting_sync1;
      counting      <= counting_sync2;
      counting_post <= counting;
   end

   always @(posedge clkvalue) begin
      if (!counting && counting_post)
	// On falling edge of counting, copy rateCtr*8 from clkref
	// clock domain to value in the clkvalue clock domain.  This
	// makes 'value' the measured frequency of clktest in 100
	// Hz. Using a left bit shift to do the *8 which compensates
	// for the use of clktest/8.
   	value <= {{32-REFCTR_SIZE-3{1'b0}}, rateCtr, 3'b0};
   end

   //
   // **** clkvalue Clock Domain ****
   
endmodule
